#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>

static const char *type_name[] = {
	"h-link",
	"dir",
	"file",
	"s-link",
	"blkdev",
	"chrdev",
	"socket",
	"fifo",
};

unsigned long __inline__ get_long(const unsigned char *p)
{
	return *(p + 0) << 24 |
	       *(p + 1) << 16 |
	       *(p + 2) << 8  |
 	       *(p + 3);
}

unsigned int output_fileinfo(unsigned char *base, unsigned int offset,int nest)
{
	unsigned long next;
	unsigned char type;
	next = get_long(base + offset);
	type = next & 0x0f;
	next &= ~0x0f;

	printf("%08x %c %6s%*c", 
	       offset, (type & 0x08)?'x':' ', type_name[type & 0x07], nest + 1, ' ');
	printf("%s : ", base + offset + 16);
	switch(type & 0x07) {
	case 0: /* Hard link */
		printf("link at %08lx\n", get_long(base + offset + 4));
		break ;
	case 1: { /* directory */
		unsigned long child;
		printf("\n");
		child = get_long(base + offset + 4);
		if (child != offset) {
			nest++;
			do {
				child = output_fileinfo(base, child, nest);
			} while(child);
		}
		break;
	}
	case 2: /* file */
		printf("size %ld byte\n", get_long(base + offset + 8));
		break ;
	case 3: { /* symlink */
		unsigned int body = 16;
		while (*(base + offset + body)) body++;
		body += 0x0f;
		body &= ~0x0f;
		printf("link at %s\n", base + offset + body);
		break ;
	}
	case 4:
	case 5: { /* dev node */
		unsigned int no;
		no = get_long(base + offset + 4);
		printf("%d,%d\n", no >> 16, no & 0xffff);
		break;
	}
	case 6:
	case 7:
		printf("\n");
		break;
	}
	return next;
}

int main(int argc, char *argv[])
{
	int fd;
	char *map;
	off_t len;
	unsigned long offset;

	if (argc != 2) {
		printf("Usage: dumpromfs romfsimage\n");
		return 1;
	}
	fd = open(argv[1], O_RDONLY);
	if (fd >= 0) {
		len = lseek(fd, 0, SEEK_END) + 1;
		map = mmap(0, len, PROT_READ, MAP_SHARED, fd, 0);
		if (memcmp("-rom1fs-", map, 8) == 0) {
			offset = 16;
			printf("Volume - %s\n", map + offset);
			while(*(map + offset) != 0) offset++;
			offset += 0x0f;
			offset &= ~0x0f;
			do {
				offset = output_fileinfo(map, offset, 0);
			} while(offset != 0);
		}
		munmap(map, len);
		close(fd);
		return 0;
	} else {
		perror("dumpromfs");
		return 0;
	}
}
